"""
dlg_example_cmpts component module.

This is the module of daliuge_component_examples containing DALiuGE data
components. Here you put your main data classes and objects.

Typically a component project will contain multiple components and will
then result in a single EAGLE palette.

Be creative! do whatever you need to do!
"""
import logging

from dlg.drop import AbstractDROP

logger = logging.getLogger(__name__)

##
# @brief MyData
# @details Template app for demonstration only!
# Replace the documentation with whatever you want/need to show in the DALiuGE
# workflow editor. The dataclass parameter should contain the relative
# Pythonpath to import MyApp.
#
# @par EAGLE_START
# @param category File
# @param appclass DropClass/dlg_example_cmpts.MyData/String/ComponentParameter/readonly//False/False/Import direction for application class
# @param dummy Dummy parameter/ /String/ApplicationArgument/readwrite//False/False/Dummy modifyable parameter
# @param dummy Dummy in//float/InputPort/readwrite//False/False/Dummy producer port
# @param dummy Dummy out//float/OutputPort/readwrite//False/False/Dummy consumer port
# @par EAGLE_END

# Data components usually directly inhert from the AbstractDROP class. Please
# refer to the Developer Guide for more information.


class MyDataDROP(AbstractDROP):
    """
    A dummy dataDROP that points to nothing.
    """

    def initialize(self, **kwargs):
        pass

    def getIO(self):
        return f"Hello from {__class__.__name__}"

    @property
    def dataURL(self):
        return "Hello from the dataURL method"



from dlg import drop
import pytest, unittest
import warnings
import os
import pickle
import urllib
import http

from glob import glob
from dlg import droputils

from dlg_example_cmpts import (
    MyBranch,
    LengthCheck,
    MyDataDROP,
    FileGlob,
    PickOne,
    String2JSON,
    ExtractColumn,
    AdvUrlRetrieve,
    GenericGather,
)
from dlg.apps.simple import RandomArrayApp
from dlg.drop import FileDROP, InMemoryDROP, NullDROP
from dlg.ddap_protocol import DROPStates
import logging
import json
import numpy as np
from time import sleep

logger = logging.Logger(__name__)

given = pytest.mark.parametrize


class TestMyApps(unittest.TestCase):
    def _runBranchTest(self, arrayDROP):
        """
        Execute the actual test given the arrayDROP on input
        """
        i = NullDROP("i", "i")  # just to be able to start the execution
        m = InMemoryDROP("m", "m")  # the receiving drop
        b = MyBranch("b", "b")  # the branch drop
        n = InMemoryDROP("n", "n")  # the memory drop for the NO branch
        y = InMemoryDROP("y", "y")  # the memory drop for the YES branch

        # connect the graph nodes
        i.addConsumer(arrayDROP)
        m.addProducer(arrayDROP)
        m.addConsumer(b)
        b.addOutput(y)
        b.addOutput(n)
        # start the graph
        with droputils.DROPWaiterCtx(self, b, timeout=1):
            i.setCompleted()

        # read the array back from the intermediate memory drop
        data = pickle.loads(droputils.allDropContents(m))
        # calculate the mean
        mean = np.mean(data)

        # check which branch should have been taken
        t = [y if mean < 0.5 else n][0]
        while t.status < 2:
            # make sure the graph has reached this point
            # status == 2 is COMPLETED, anything above is not expected
            sleep(0.001)
        # load the mean from the correct branch memory drop
        res = pickle.loads(droputils.allDropContents(t))
        # and check whether it is the same as the calculated one
        return (t.oid, res, mean)

    def test_myBranch_class(self):
        """
        Test creates two random arrays in memory drops, one with a
        mean below and the other above 0.5. It runs two graphs against
        each of the arrays drops and checks whether the branch is
        traversed on the correct side. It also checks whether the
        derived values are correct.
        """
        # create and configure the creation of the random array.
        l = RandomArrayApp("l", "l")
        l.integer = False
        l.high = 0.5
        (oid, resLow, meanLow) = self._runBranchTest(l)
        self.assertEqual(oid, "y")
        self.assertEqual(resLow, meanLow)

        h = RandomArrayApp("h", "h")
        h.integer = False
        h.low = 0.5
        h.high = 1
        (oid, resHigh, meanHigh) = self._runBranchTest(h)
        self.assertEqual(oid, "n")
        self.assertEqual(resHigh, meanHigh)

    def _runLengthTest(self, arrayDROP):
        """
        Execute the actual test given the arrayDROP on input
        """
        b = LengthCheck("b", "b")  # the branch drop
        n = InMemoryDROP("n", "n")  # the memory drop for the NO branch
        y = InMemoryDROP("y", "y")  # the memory drop for the YES branch
        if not isinstance(arrayDROP, InMemoryDROP):
            startDrop = NullDROP(
                "i", "i"
            )  # just to be able to start the execution
            m = InMemoryDROP("m", "m")  # the receiving drop
            startDrop.addConsumer(arrayDROP)
            m.addProducer(arrayDROP)
            m.addConsumer(b)
        else:
            startDrop = m = arrayDROP
            startDrop.addConsumer(b)

        # connect the Y and N nodes
        b.addOutput(y)
        b.addOutput(n)
        # start the graph
        with droputils.DROPWaiterCtx(self, b, timeout=30):
            startDrop.setCompleted()
        # read the array back from the intermediate memory drop
        data = pickle.loads(droputils.allDropContents(m))
        # calculate the length
        if isinstance(data, list):
            data = np.array(data)
        if isinstance(data, np.ndarray):
            length = data.size
            t = [n if length < 1 else y][0]
            while t.status < 2:
                # make sure the graph has reached this point
                # status == 2 is COMPLETED, anything above is not expected
                sleep(0.001)
            # load the mean from the correct branch memory drop
            dummy = pickle.loads(droputils.allDropContents(t))
            res = dummy.size
            # and check whether it is the same as the calculated one
            return (t.oid, res, length)
        else:
            length = None
            return (None, None, None)

    def test_LengthCheck_class(self):
        """
        Test creates two random arrays in memory drops, one with zero length,
        the other with 100. It runs two graphs against
        each of the arrays and checks whether the branch is
        traversed on the correct side.
        """
        # check > 0 size (default == 100)
        h = RandomArrayApp("h", "h")
        h.integer = False
        (oid, check, length) = self._runLengthTest(h)
        self.assertEqual(oid, "y")
        self.assertEqual(check, length)

    def test_LengthCheck_list(self):
        """
        Test creates two random arrays in memory drops, one with zero length,
        the other with 100. It runs two graphs against
        each of the arrays and checks whether the branch is
        traversed on the correct side.
        """
        # check > 0 size (default == 100)
        h = InMemoryDROP("h", "h")
        h.write(pickle.dumps([1, 2, 3]))
        (oid, check, length) = self._runLengthTest(h)
        self.assertEqual(oid, "y")
        self.assertEqual(check, length)

    def test_LengthCheck_class_zero(self):
        """
        Test creates a zero length array and checks whether the branch is
        traversed on the correct side.
        """

        # check zero size
        l = RandomArrayApp("l", "l")
        l.integer = False
        l.size = 0
        (oid, check, length) = self._runLengthTest(l)
        self.assertEqual(oid, "n")
        self.assertEqual(check, length)

    @pytest.mark.filterwarnings(
        "ignore::pytest.PytestUnhandledThreadExceptionWarning"
    )
    def test_LengthCheck_wrongType(self):
        # check wrong type
        w = InMemoryDROP("w", "w")
        b = LengthCheck("b", "b")  # the branch drop
        n = InMemoryDROP("n", "n")  # the memory drop for the NO branch
        y = InMemoryDROP("y", "y")  # the memory drop for the YES branch
        w.addConsumer(b)
        n.addProducer(b)
        y.addProducer(b)
        w.write(pickle.dumps(b"abcdef"))
        with droputils.DROPWaiterCtx(self, b, timeout=10):
            w.setCompleted()
        self.assertRaises(TypeError, b.readData)

    def test_LengthCheck_scalar(self):
        # check array scalar
        w = InMemoryDROP("w", "w")
        w.write(pickle.dumps(np.array(1)))
        (oid, check, length) = self._runLengthTest(w)
        self.assertEqual(check, length)

    def test_FileGlob_class(self):
        """
        Testing the globbing method finding *this* file
        """
        i = NullDROP("i", "i")  # just to be able to start the execution
        g = FileGlob("g", "g")
        m = InMemoryDROP("m", "m")
        g.addInput(i)
        m.addProducer(g)
        g.wildcard = os.path.basename(os.path.realpath(__file__))
        g.filepath = os.path.dirname(os.path.realpath(__file__))
        fileList = glob(f"{g.filepath}/{g.wildcard}")
        with droputils.DROPWaiterCtx(self, m, timeout=10):
            i.setCompleted()
        res = pickle.loads(droputils.allDropContents(m))
        self.assertEqual(fileList, res)

    def test_FileGlob_class_nomatch(self):
        """
        Testing the globbing method finding *this* file
        """
        i = NullDROP("i", "i")  # just to be able to start the execution
        g = FileGlob("g", "g")
        m = InMemoryDROP("m", "m")
        g.addInput(i)
        m.addProducer(g)
        g.wildcard = os.path.basename("rubbish")
        g.filepath = os.path.dirname(os.path.realpath(__file__))
        fileList = glob(f"{g.filepath}/{g.wildcard}")
        with droputils.DROPWaiterCtx(self, m, timeout=10):
            i.setCompleted()
        res = pickle.loads(droputils.allDropContents(m))
        self.assertEqual(fileList, res)

    def test_PickOne_class(self):
        """
        Testing the PickOne app
        """
        i = NullDROP("i", "i")  # just to be able to start the execution
        l = RandomArrayApp("l", "l")
        l.integer = True
        l.high = 100
        l.size = 10
        a = InMemoryDROP("a", "a", type="array", nm="start_array")
        p = PickOne("p", "p")
        r = InMemoryDROP("r", "r", type="array", nm="rest_array")
        o = InMemoryDROP("o", "o", type="scalar", nm="value")

        i.addConsumer(l)
        a.addProducer(l)
        p.addInput(a)
        r.addProducer(p)  # this should contain the rest elements
        o.addProducer(p)  # this should contain the first element
        with droputils.DROPWaiterCtx(self, o, timeout=10):
            i.setCompleted()
        in_array = pickle.loads(droputils.allDropContents(a))
        first = pickle.loads(droputils.allDropContents(o))
        rest_array = pickle.loads(droputils.allDropContents(r))
        self.assertEqual(in_array[0], first)
        self.assertEqual(all(in_array[1:]), all(rest_array))

    def test_PickOne_wrong_input_type(self):
        """
        Testing the PickOne with wrong input type
        """
        a = InMemoryDROP("a", "a", type="array", nm="start_array")
        a.write(pickle.dumps(1))  # this is scalar not array
        p = PickOne("p", "p")
        a.addConsumer(p)
        with self.assertRaises(TypeError):
            a.setCompleted()
            p.readData()

    def test_PickOne_0dim(self):
        """
        Testing the PickOne with 0dim input
        """
        a = InMemoryDROP("a", "a")
        a.write(pickle.dumps(np.array(1)))  # this is 0dim not array
        a.name = "start_array"
        p = PickOne("p", "p")

        p.addInput(a)
        with droputils.DROPWaiterCtx(self, p, timeout=10):
            a.setCompleted()
        self.assertRaises(TypeError)

    def test_String2JSON(self):
        """
        Testing String2JSON with correct input
        """
        example = b'[{"a":1, "b":2},[1,2,3,4,5]]'
        a = InMemoryDROP("a", "a")
        a.name = "string"
        a.write(example)
        # f = FileDROP('f', 'f', filepath='/tmp/dlg/workspace/NGASLogProcess11_2021-12-16T07-50-33.653962/2021-12-16T07_48_09_-3_0')
        p = String2JSON("p", "p")
        o = InMemoryDROP("o", "o")

        p.addInput(a)
        p.addOutput(o)
        with droputils.DROPWaiterCtx(self, o, timeout=10):
            a.setCompleted()

        example = json.loads(droputils.allDropContents(a))
        result = pickle.loads(droputils.allDropContents(o))
        self.assertEqual(result, example)

    def test_String2JSON_wrong_type(self):
        """
        Testing String2JSON with wrong input
        """
        example = b"gibberish"
        a = InMemoryDROP("a", "a")
        a.write(example)
        a.name = "string"
        p = String2JSON("p", "p")
        o = InMemoryDROP("o", "o")

        p.addInput(a)
        p.addOutput(o)
        with droputils.DROPWaiterCtx(self, o, timeout=10):
            a.setCompleted()
        self.assertRaises(TypeError)

    def test_ExtractColumn(self):
        """
        Testing ExtractColumn
        """
        table_array = np.arange(300).reshape(100, 3)
        column = table_array[:, 1]  # select 1st column
        a = InMemoryDROP("a", "a")
        a.write(pickle.dumps(table_array))
        a.name = "table_array"
        e = ExtractColumn("e", "e")
        e.index = 1
        o = InMemoryDROP("o", "o")
        o.name = "column"

        e.addInput(a)
        e.addOutput(o)
        with droputils.DROPWaiterCtx(self, o, timeout=10):
            a.setCompleted()
        output = pickle.loads(droputils.allDropContents(o))
        self.assertListEqual(list(column), list(output))

    def test_ExtractColumn_wrongShape(self):
        """
        Testing ExtractColumn
        """
        table_array = np.arange(300)
        a = InMemoryDROP("a", "a")
        a.write(pickle.dumps(table_array))
        a.name = "table_array"
        e = ExtractColumn("e", "e")
        e.index = 1
        o = InMemoryDROP("o", "o")
        o.name = "column"

        e.addInput(a)
        e.addOutput(o)
        with droputils.DROPWaiterCtx(self, o, timeout=10):
            a.setCompleted()
        self.assertRaises(TypeError)

    def test_AdvUrlRetrieve(self):
        """
        Testing AdvUrlRetrieve
        """
        testContent = {"args": {"daliuge": "great"}}
        testUrl = "https://httpbin.org/get?daliuge=%i0"
        testPart = "great"
        a = InMemoryDROP("a", "a")
        a.write(pickle.dumps(testPart))
        a.name = "partUrl"
        e = AdvUrlRetrieve("e", "e")
        e.urlTempl = testUrl
        o = InMemoryDROP("o", "o")
        o.name = "content"

        e.addInput(a)
        e.addOutput(o)
        with droputils.DROPWaiterCtx(self, o, timeout=10):
            a.setCompleted()
        content = json.loads(droputils.allDropContents(o))
        self.assertEqual(content["args"], testContent["args"])

    def test_AdvUrlRetrieve_wrongUrl(self):
        """
        Testing AdvUrlRetrieve with wrong URL
        """
        testContent = {"args": {"daliuge": "great"}}
        testUrl = "https://dummy/get?daliuge=%i0"
        testPart = "great"
        a = InMemoryDROP("a", "a")
        a.write(pickle.dumps(testPart))
        a.name = "partUrl"
        e = AdvUrlRetrieve("e", "e")
        e.urlTempl = testUrl
        o = InMemoryDROP("o", "o")
        o.name = "content"

        e.addInput(a)
        e.addOutput(o)
        with droputils.DROPWaiterCtx(self, o, timeout=10):
            a.setCompleted()
        self.assertRaises(urllib.error.URLError)

    def test_AdvUrlRetrieve_invalidUrl(self):
        """
        Testing AdvUrlRetrieve with invalid URL
        """
        testContent = {"args": {"daliuge": "great"}}
        testUrl = "https://dummy\ get?daliuge=%i0"
        testPart = "great"
        a = InMemoryDROP("a", "a")
        a.write(pickle.dumps(testPart))
        a.name = "partUrl"
        e = AdvUrlRetrieve("e", "e")
        e.urlTempl = testUrl
        o = InMemoryDROP("o", "o")
        o.name = "content"

        e.addInput(a)
        e.addOutput(o)
        with droputils.DROPWaiterCtx(self, o, timeout=10):
            a.setCompleted()
        # content = json.loads(pickle.loads(droputils.allDropContents(o)))
        # self.assertEqual(content["args"], testContent["args"])
        self.assertRaises(http.client.InvalidURL)

    def test_AdvUrlRetrieve_noOutput(self):
        """
        Testing AdvUrlRetrieve without output
        """
        testContent = {"args": {"daliuge": "great"}}
        testUrl = "https://httpbin.org/get?daliuge=%i0"
        testPart = "great"
        a = InMemoryDROP("a", "a")
        a.write(pickle.dumps(testPart))
        a.name = "partUrl"
        e = AdvUrlRetrieve("e", "e")
        e.urlTempl = testUrl
        # o = InMemoryDROP("o", "o")
        # o.name = "content"

        e.addInput(a)
        # e.addOutput(o)
        with droputils.DROPWaiterCtx(self, e, timeout=10):
            a.setCompleted()
        # content = json.loads(pickle.loads(droputils.allDropContents(a)))
        # self.assertEqual(content["args"], testContent["args"])
        with self.assertRaisesRegex(Exception, "At least one output required"):
            raise Exception("At least one output required")

    def test_AdvUrlRetrieve_wrongType(self):
        """
        Testing AdvUrlRetrieve with malicious type
        """
        testContent = {"args": {"daliuge": "great"}}
        testUrl = "https://httpbin.org/get?daliuge=%i0"
        testPart = b"0123456"
        a = InMemoryDROP("a", "a")
        a.write(pickle.dumps(testPart))
        a.name = "partUrl"
        e = AdvUrlRetrieve("e", "e")
        e.urlTempl = testUrl
        o = InMemoryDROP("o", "o")
        o.name = "content"

        e.addInput(a)
        e.addOutput(o)
        with droputils.DROPWaiterCtx(self, o, timeout=10):
            a.setCompleted()
        self.assertRaises(TypeError)

    def test_AdvUrlRetrieve_wrongOutput(self):
        """
        Testing AdvUrlRetrieve with wrong output name
        """
        testContent = {"args": {"daliuge": "great"}}
        testUrl = "https://httpbin.org/get?daliuge=%i0"
        testPart = "great"
        a = InMemoryDROP("a", "a")
        a.write(pickle.dumps(testPart))
        a.name = "partUrl"
        e = AdvUrlRetrieve("e", "e")
        e.urlTempl = testUrl
        o = InMemoryDROP("o", "o")
        o.name = "wrongName"

        e.addInput(a)
        e.addOutput(o)
        with droputils.DROPWaiterCtx(self, o, timeout=10):
            a.setCompleted()
        self.assertRaises(TypeError)

    def test_GenericGather(self):
        """
        Testing the simple GenericGather app
        """
        a = InMemoryDROP("a", "a")
        b = InMemoryDROP("b", "b")
        o = InMemoryDROP("o", "o")
        g = GenericGather("g", "g")
        a.write((b"a" * 10))
        a.len = 10
        b.write((b"b" * 10))
        b.len = 10
        g.addInput(a)
        g.addInput(b)
        g.addOutput(o)
        with droputils.DROPWaiterCtx(self, o, timeout=10):
            a.setCompleted()
            b.setCompleted()
        content = content = droputils.allDropContents(o)
        self.assertEqual(content, b"a" * 10 + b"b" * 10)

    def test_myData_class(self):
        """
        Dummy getIO method test for data drop
        """
        assert MyDataDROP("a", "a").getIO() == "Hello from MyDataDROP"

    def test_myData_dataURL(self):
        """
        Dummy dataURL method test for data drop
        """
        assert MyDataDROP("a", "a").dataURL == "Hello from the dataURL method"



import pytest, unittest
import os
import pickle
import urllib
import http

from glob import glob
from dlg import droputils

from dlg_example_funcs import simple
from dlg.apps.pyfunc import PyFuncApp
from dlg.drop import FileDROP, InMemoryDROP, NullDROP
from dlg.ddap_protocol import DROPStates
import logging
import json

logger = logging.Logger(__name__)

given = pytest.mark.parametrize


class TestMyFuncs(unittest.TestCase):
    def test_output(self):
        result = simple.output("World", kw="")
        self.assertEqual(result, "Hello World and ")



"""Distributed Fourier Transform Module."""

import numpy
import itertools

from crocodile.synthesis import (
    fft,
    ifft,
    pad_mid,
    extract_mid,
)

def fmt(x):
    """
    :param x: x

    :return: x
    """
    if x >= 1024 * 1024 and (x % (1024 * 1024)) == 0:
        return "%dM" % (x // 1024 // 1024)
    if x >= 1024 and (x % 1024) == 0:
        return "%dk" % (x // 1024)
    return "%d" % x


def mark_range(
    lbl, x0, x1=None, y0=None, y1=None, ax=None, x_offset=1 / 200, linestyle="--"
):
    """Helper for marking ranges in a graph.

    :param lbl: x
    :param x0: x
    :param x1: x
    :param y1: x
    :param ax: x
    :param x_offset: x
    :param linestyle: linestyle

    """
    if ax is None:
        ax = pylab.gca()
    if y0 is None:
        y0 = ax.get_ylim()[1]
    if y1 is None:
        y1 = ax.get_ylim()[0]
    wdt = ax.get_xlim()[1] - ax.get_xlim()[0]
    ax.add_patch(
        patches.PathPatch(patches.Path([(x0, y0), (x0, y1)]), linestyle=linestyle)
    )
    if x1 is not None:
        ax.add_patch(
            patches.PathPatch(patches.Path([(x1, y0), (x1, y1)]), linestyle=linestyle)
        )
    else:
        x1 = x0
    if pylab.gca().get_yscale() == "linear":
        lbl_y = (y0 * 7 + y1) / 8
    else:
        # Some type of log scale
        lbl_y = (y0 ** 7 * y1) ** (1 / 8)
    ax.annotate(lbl, (x1 + x_offset * wdt, lbl_y))


def find_x_sorted_smooth(xs, ys, y):
    """Find sorted smooth.

    :param xs: x
    :param ys: x
    :param y: x

    :return: xs

    """

    assert len(xs) == len(ys)
    pos = numpy.searchsorted(ys, y)
    if pos <= 0:
        return xs[0]
    if pos >= len(ys) or ys[pos] == ys[pos - 1]:
        return xs[len(ys) - 1]
    w = (y - ys[pos - 1]) / (ys[pos] - ys[pos - 1])
    return xs[pos - 1] * (1 - w) + xs[pos] * w


def find_x_sorted_logsmooth(xs, ys, y):
    """Find sorted log smooth.

    :param xs: x
    :param ys: x
    :param y: x

    :return: log xs

    """
    return find_x_sorted_smooth(xs, numpy.log(numpy.maximum(1e-100, ys)), numpy.log(y))


def whole(xs):
    """."""
    return numpy.all(numpy.abs(xs - numpy.around(xs)) < 1e-13)


def make_subgrid_and_facet(
    G,
    nsubgrid,
    xA_size,
    subgrid_A,
    subgrid_off,
    nfacet,
    yB_size,
    facet_B,
    facet_off,
):
    """

    Calculate the actual subgrids & facets

    :param G: x
    :param nsubgrid: x
    :param xA_size: x
    :param subgrid_A: x
    :param subgrid_off: x
    :param nfacet: x
    :param yB_size: yB_size
    :param facet_B: facet_B
    :param facet_off: facet_off

    :return: subbrig and facet

    """
    FG = fft(G)
    subgrid = numpy.empty((nsubgrid, xA_size), dtype=complex)
    for i in range(nsubgrid):
        subgrid[i] = subgrid_A[i] * extract_mid(numpy.roll(G, -subgrid_off[i]), xA_size)
    facet = numpy.empty((nfacet, yB_size), dtype=complex)
    for j in range(nfacet):
        facet[j] = facet_B[j] * extract_mid(numpy.roll(FG, -facet_off[j]), yB_size)
    return subgrid, facet


def facets_to_subgrid_1(
    facet,
    nsubgrid,
    nfacet,
    xM_yN_size,
    Fb,
    Fn,
    yP_size,
    facet_m0_trunc,
    subgrid_off,
    N,
    xMxN_yP_size,
    xN_yP_size,
    xM_yP_size,
    dtype,
):
    """

    Facet to subgrid 1.

    param facet:
    param nsubgrid:
    param nfacet:
    param xM_yN_size:
    param Fb:
    param Fn:
    param yP_size:
    param facet_m0_trunc:
    param subgrid_off:
    param N:
    param xMxN_yP_size:
    param xN_yP_size:
    param xM_yP_size:
    param dtype:

    return: RNjMiBjFj

    """
    RNjMiBjFj = numpy.empty((nsubgrid, nfacet, xM_yN_size), dtype=dtype)
    for j in range(nfacet):
        BjFj = ifft(pad_mid(facet[j] * Fb, yP_size))
        for i in range(nsubgrid):
            # TODO Seperate into a different function
            MiBjFj = facet_m0_trunc * extract_mid(
                numpy.roll(BjFj, -subgrid_off[i] * yP_size // N), xMxN_yP_size
            )
            MiBjFj_sum = numpy.array(extract_mid(MiBjFj, xM_yP_size))
            MiBjFj_sum[: xN_yP_size // 2] += MiBjFj[-xN_yP_size // 2 :]
            MiBjFj_sum[-xN_yP_size // 2 :] += MiBjFj[: xN_yP_size // 2 :]
            RNjMiBjFj[i, j] = Fn * extract_mid(fft(MiBjFj_sum), xM_yN_size)
    return RNjMiBjFj


def facets_to_subgrid_2(nmbfs, i, xM_size, nfacet, facet_off, N, subgrid_A, xA_size):
    """

    Facet to subgrid 2.
    param nmbfs:
    param i:
    param xM_size:
    param nfacet:
    param facet_off:
    param N:
    param subgrid_A:
    param xA_size:

    return:
    """
    approx = numpy.zeros(xM_size, dtype=complex)
    for j in range(nfacet):
        approx += numpy.roll(pad_mid(nmbfs[i, j], xM_size), facet_off[j] * xM_size // N)
    return subgrid_A[i] * extract_mid(ifft(approx), xA_size)


def subgrid_to_facet_1(
    subgrid, nsubgrid, nfacet, xM_yN_size, xM_size, facet_off, N, Fn
):
    """

    param subgrid:
    param nsubgrid:
    param nfacet:
    param xM_yN_size:
    param xM_size:
    param facet_off:
    param N:
    param Fn:

    return:
    """
    FNjSi = numpy.empty((nsubgrid, nfacet, xM_yN_size), dtype=complex)
    for i in range(nsubgrid):
        FSi = fft(pad_mid(subgrid[i], xM_size))
        for j in range(nfacet):
            FNjSi[i, j] = extract_mid(
                numpy.roll(FSi, -facet_off[j] * xM_size // N), xM_yN_size
            )
    return Fn * FNjSi


def subgrid_to_facet_2(
    nafs,
    j,
    yB_size,
    nsubgrid,
    xMxN_yP_size,
    xM_yP_size,
    xN_yP_size,
    facet_m0_trunc,
    yP_size,
    subgrid_off,
    N,
    Fb,
    facet_B,
):
    """

    param nafs:
    param j:
    param yB_size:
    param nsubgrid:
    param xMxN_yP_size:
    param xM_yP_size:
    param xN_yP_size:
    param facet_m0_trunc:
    param yP_size:
    param subgrid_off:
    param N:
    param Fb:
    param facet_B:

    return:
    """
    approx = numpy.zeros(yB_size, dtype=complex)
    for i in range(nsubgrid):

        #TODO Seperate function
        NjSi = numpy.zeros(xMxN_yP_size, dtype=complex)
        NjSi_mid = extract_mid(NjSi, xM_yP_size)
        NjSi_mid[:] = ifft(
            pad_mid(nafs[i, j], xM_yP_size)
        )  # updates NjSi via reference!
        NjSi[-xN_yP_size // 2 :] = NjSi_mid[: xN_yP_size // 2]
        NjSi[: xN_yP_size // 2 :] = NjSi_mid[-xN_yP_size // 2 :]
        FMiNjSi = fft(
            numpy.roll(
                pad_mid(facet_m0_trunc * NjSi, yP_size), subgrid_off[i] * yP_size // N
            )
        )
        approx += extract_mid(FMiNjSi, yB_size)
    return approx * Fb * facet_B[j]


# Given that the amount of data has been squared, performance is a bit more of a concern now. Fortunately, the entire procedure is completely separable, so let us first re-define the operations to work on one array axis exclusively:
def slice_a(fill_val, axis_val, dims, axis):
    """

    param fill_val:
    param axis_val:
    param dims:
    param axis:

    return:
    """
    return tuple([axis_val if i == axis else fill_val for i in range(dims)])


def pad_mid_a(a, N, axis):
    """

    param a:
    param N:
    param axis:

    return:
    """
    N0 = a.shape[axis]
    if N == N0:
        return a
    pad = slice_a(
        (0, 0), (N // 2 - N0 // 2, (N + 1) // 2 - (N0 + 1) // 2), len(a.shape), axis
    )
    return numpy.pad(a, pad, mode="constant", constant_values=0.0)


def extract_mid_a(a, N, axis):
    """

    param a:
    param N:
    param axis:

    return:
    """
    assert N <= a.shape[axis]
    cx = a.shape[axis] // 2
    if N % 2 != 0:
        slc = slice(cx - N // 2, cx + N // 2 + 1)
    else:
        slc = slice(cx - N // 2, cx + N // 2)
    return a[slice_a(slice(None), slc, len(a.shape), axis)]


def fft_a(a, axis):
    """

    param a:
    param axis:

    return:
    """
    return numpy.fft.fftshift(
        numpy.fft.fft(numpy.fft.ifftshift(a, axis), axis=axis), axis
    )


def ifft_a(a, axis):

    """

    param a:
    param axis:

    return:
    """
    return numpy.fft.fftshift(
        numpy.fft.ifft(numpy.fft.ifftshift(a, axis), axis=axis), axis
    )


def broadcast_a(a, dims, axis):
    """

    param a:
    param dims:
    param axis:

    return:
    """
    slc = [numpy.newaxis] * dims
    slc[axis] = slice(None)
    return a[slc]


def broadcast_a(a, dims, axis):
    """

    param a:
    param dims:
    param axis:
    return:
    """
    return a[slice_a(numpy.newaxis, slice(None), dims, axis)]


# This allows us to define the two fundamental operations - going from $F$ to $b\ast F$ and from $b\ast F$ to $n\ast m(b\ast F)$ separately:


def prepare_facet(facet, axis, Fb, yP_size):
    """

    param facet:
    param axis:
    param Fb:
    param yP_size:
    return:
    """
    BF = pad_mid_a(facet * broadcast_a(Fb, len(facet.shape), axis), yP_size, axis)
    BF = ifft_a(BF, axis)
    return BF


def extract_subgrid(
    BF,
    i,
    axis,
    subgrid_off,
    yP_size,
    xMxN_yP_size,
    facet_m0_trunc,
    xM_yP_size,
    Fn,
    xM_yN_size,
    N,
):
    """

    param BF:
    param i:
    param axis:
    param subgrid_off:
    param yP_size:
    param xMxN_yP_size:
    param facet_m0_trunc:
    param xM_yP_size:
    param Fn:
    param xM_yN_size:
    param N:
    return:
    """
    dims = len(BF.shape)
    BF_mid = extract_mid_a(
        numpy.roll(BF, -subgrid_off[i] * yP_size // N, axis), xMxN_yP_size, axis
    )
    MBF = broadcast_a(facet_m0_trunc, dims, axis) * BF_mid
    MBF_sum = numpy.array(extract_mid_a(MBF, xM_yP_size, axis))
    xN_yP_size = xMxN_yP_size - xM_yP_size
    # [:xN_yP_size//2] / [-xN_yP_size//2:] for axis, [:] otherwise
    slc1 = slice_a(slice(None), slice(xN_yP_size // 2), dims, axis)
    slc2 = slice_a(slice(None), slice(-xN_yP_size // 2, None), dims, axis)
    MBF_sum[slc1] += MBF[slc2]
    MBF_sum[slc2] += MBF[slc1]
    return broadcast_a(Fn, len(BF.shape), axis) * extract_mid_a(
        fft_a(MBF_sum, axis), xM_yN_size, axis
    )


# TODO: Should this be here?
# @interact(xs=(0, N), ys=(0, N))
def test_accuracy(
    nsubgrid,
    xA_size,
    nfacet,
    yB_size,
    N,
    subgrid_off,
    subgrid_A,
    facet_off,
    facet_B,
    xM_yN_size,
    xM_size,
    Fb,
    yP_size,
    xMxN_yP_size,
    facet_m0_trunc,
    xM_yP_size,
    Fn,
    xs=252,
    ys=252,
):
    """

    param nsubgrid:
    param xA_size:
    param nfacet:
    param yB_size:
    param N:
    param subgrid_off:
    param subgrid_A:
    param facet_off:
    param facet_B:
    param xM_yN_size:
    param xM_size:
    param Fb:
    param yP_size:
    param xMxN_yP_size:
    param facet_m0_trunc:
    param xM_yP_size:
    param Fn:
    param xs:
    param ys:
    """
    subgrid_2 = numpy.empty((nsubgrid, nsubgrid, xA_size, xA_size), dtype=complex)
    facet_2 = numpy.empty((nfacet, nfacet, yB_size, yB_size), dtype=complex)

    # G_2 = numpy.exp(2j*numpy.pi*numpy.random.rand(N,N))*numpy.random.rand(N,N)/2
    # FG_2 = fft(G_2)

    FG_2 = numpy.zeros((N, N))
    FG_2[ys, xs] = 1
    G_2 = ifft(FG_2)

    for i0, i1 in itertools.product(range(nsubgrid), range(nsubgrid)):
        subgrid_2[i0, i1] = extract_mid(
            numpy.roll(G_2, (-subgrid_off[i0], -subgrid_off[i1]), (0, 1)), xA_size
        )
        subgrid_2[i0, i1] *= numpy.outer(subgrid_A[i0], subgrid_A[i1])
    for j0, j1 in itertools.product(range(nfacet), range(nfacet)):
        facet_2[j0, j1] = extract_mid(
            numpy.roll(FG_2, (-facet_off[j0], -facet_off[j1]), (0, 1)), yB_size
        )
        facet_2[j0, j1] *= numpy.outer(facet_B[j0], facet_B[j1])

    NMBF_NMBF = numpy.empty(
        (nsubgrid, nsubgrid, nfacet, nfacet, xM_yN_size, xM_yN_size), dtype=complex
    )
    for j0, j1 in itertools.product(range(nfacet), range(nfacet)):
        BF_F = prepare_facet(facet_2[j0, j1], 0, Fb, yP_size)
        BF_BF = prepare_facet(BF_F, 1, Fb, yP_size)
        for i0 in range(nsubgrid):
            NMBF_BF = extract_subgrid(
                BF_BF,
                i0,
                0,
                subgrid_off,
                yP_size,
                xMxN_yP_size,
                facet_m0_trunc,
                xM_yP_size,
                Fn,
                xM_yN_size,
                N,
            )
            for i1 in range(nsubgrid):
                NMBF_NMBF[i0, i1, j0, j1] = extract_subgrid(
                    NMBF_BF,
                    i1,
                    1,
                    subgrid_off,
                    yP_size,
                    xMxN_yP_size,
                    facet_m0_trunc,
                    xM_yP_size,
                    Fn,
                    xM_yN_size,
                    N,
                )

    pylab.rcParams["figure.figsize"] = 16, 8
    err_mean = err_mean_img = 0
    for i0, i1 in itertools.product(range(nsubgrid), range(nsubgrid)):
        approx = numpy.zeros((xM_size, xM_size), dtype=complex)
        for j0, j1 in itertools.product(range(nfacet), range(nfacet)):
            approx += numpy.roll(
                pad_mid(NMBF_NMBF[i0, i1, j0, j1], xM_size),
                (facet_off[j0] * xM_size // N, facet_off[j1] * xM_size // N),
                (0, 1),
            )
        approx = extract_mid(ifft(approx), xA_size)
        approx *= numpy.outer(subgrid_A[i0], subgrid_A[i1])
        err_mean += numpy.abs(approx - subgrid_2[i0, i1]) ** 2 / nsubgrid ** 2
        err_mean_img += numpy.abs(fft(approx - subgrid_2[i0, i1])) ** 2 / nsubgrid ** 2
    # pylab.imshow(numpy.log(numpy.sqrt(err_mean)) / numpy.log(10)); pylab.colorbar(); pylab.show()
    pylab.imshow(numpy.log(numpy.sqrt(err_mean_img)) / numpy.log(10))
    pylab.colorbar()
    pylab.show()
    print(
        "RMSE:",
        numpy.sqrt(numpy.mean(err_mean)),
        "(image:",
        numpy.sqrt(numpy.mean(err_mean_img)),
        ")",
    )


"""
dlg_example_cmpts appComponent module.

This is the module of dlg_example_cmpts containing DALiuGE
application components.
Here you put your main application classes and objects.

Typically a component project will contain multiple components and will
then result in a single EAGLE palette.

Be creative! do whatever you need to do!
"""

__version__ = "0.1.17"
import json
import logging
import os
import pickle
import urllib
from glob import glob

import numpy as np
from dlg import droputils
from dlg.drop import BarrierAppDROP

try:
    from dlg.drop import BranchAppDrop
except ImportError:
    from dlg.apps.branch import BranchAppDrop

from dlg.meta import dlg_string_param

logger = logging.getLogger(__name__)

##
# @brief MyBranch
# @details Simple app to demonstrate how to write a branch actually making a
# decision and passing data on.
# Most of the code is boilerplate and can be copied verbatim. Note that a
# branch app is allowed
# to have multiple inputs, but just exactly two outputs. This example is using
# just a single input. There is an associated logical graph available on
# github:
#
#    https://github.com//EAGLE-graph-repo/examples/branchDemo.graph
#
# The application assumes to receive a random floating point array with values
# in the range [0,1] on input. It will calculate the mean of that array and
# then branch depending on whether the mean is smaller or larger than 0.5.
#
# @par EAGLE_START
# @param category PythonApp
# @param appclass Application class/dlg_example_cmpts.apps.MyBranch/String/ComponentParameter/readonly//False/False/Application class # noqa: E501
# @param execution_time Execution Time/5/Float/ComponentParameter/readonly//False/False/Estimated execution time # noqa: E501
# @param num_cpus No. of CPUs/1/Integer/ComponentParameter/readonly//False/False/Number of cores used # noqa: E501
# @param array Array//Object.Array/InputPort/readwrite//False/False/Port receiving the input array
# @param Y Y//float/OutputPort/readwrite//False/False/English Port carrying the mean value of the array if mean < 0.5
# @param N N//float/OutputPort/readwrite//False/False/English Port carrying the mean value of the array if mean >= 0.5
# @par EAGLE_END


class MyBranch(BranchAppDrop):
    def initialize(self, **kwargs):
        BranchAppDrop.initialize(self, **kwargs)

    def run(self):
        """
        Just reading the input array and calculating the mean.
        """
        input = self.inputs[0]
        data = pickle.loads(droputils.allDropContents(input))
        self.value = data.mean()

    def writeData(self):
        """
        Prepare the data and write to port (self.ind) identified by condition.
        """
        output = self.outputs[self.ind]
        d = pickle.dumps(self.value)
        output.len = len(d)
        logger.info(f">>>>>>> Writing value {self.value} to output {self.ind}")
        output.write(d)

    def condition(self):
        """
        Check value, call write method and return boolean.
        """
        if self.value < 0.5:
            self.ind = 0
            result = True
        else:
            self.ind = 1
            result = False
        self.writeData()
        return result


##
# @brief LengthCheck
#
# This branch returns true if the length of the input array is > 0. It also
# passes the array on to the output ports (will be an empty array in the case
# of N==0).
#
# @par EAGLE_START
# @param category PythonApp
# @param appclass Application class/dlg_example_cmpts.apps.MyBranch/String/ComponentParameter/readonly//False/False/Application class # noqa: E501
# @param execution_time Execution Time/5/Float/ComponentParameter/readonly//False/False/Estimated execution time # noqa: E501
# @param num_cpus No. of CPUs/1/Integer/ComponentParameter/readonly//False/False/Number of cores used # noqa: E501
# @param array Array//Object.Array/InputPort/readwrite//False/False/Port receiving the input array
# @param Y Y//float/OutputPort/readwrite//False/False/English Port carrying the mean value of the array if mean < 0.5
# @param N N//float/OutputPort/readwrite//False/False/English Port carrying the mean value of the array if mean >= 0.5
# @par EAGLE_END


class LengthCheck(BranchAppDrop):
    def initialize(self, **kwargs):
        BranchAppDrop.initialize(self, **kwargs)

    def readData(self):
        """
        Just reading the inoput array and determining the size
        """
        input = self.inputs[0]
        self.raw = droputils.allDropContents(input)
        data = pickle.loads(self.raw)
        # make sure we always have a ndarray with at least 1dim.
        if isinstance(data, np.ndarray):
            # that's what we want..
            if data.size == 0:
                # fix this strange case
                data = np.array([data])
                self.raw = pickle.dumps(data)
        elif not isinstance(data, np.ndarray) and type(data) not in (
            list,
            tuple,
        ):
            raise TypeError
        else:
            # can only be list or tuple. convert those.
            data = np.array(data)
            self.raw = pickle.dumps(data)

        self.value = data.size

    def writeData(self):
        """
        Write unmodified data to port (self.ind) identified by condition.
        """
        output = self.outputs[self.ind]
        output.len = self.value
        output.write(self.raw)

    def condition(self):
        """
        Check value, call write method and return boolean.
        """
        self.readData()
        if self.value > 0:
            self.ind = 0
            result = True
        else:
            self.ind = 1
            result = False
        self.writeData()
        return result

    def run(self):
        self.readData()


##
# @brief FileGlob
# @details An App that uses glob to find all files matching a
# template given by a filepath and a wildcard string
#
# @par EAGLE_START
# @param category PythonApp
# @param appclass Application class/dlg_example_cmpts.apps.FileGlob/String/ComponentParameter/readonly//False/False/Application class # noqa: E501
# @param execution_time Execution Time/5/Float/ComponentParameter/readonly//False/False/Estimated execution time # noqa: E501
# @param num_cpus No. of CPUs/1/Integer/ComponentParameter/readonly//False/False/Number of cores used # noqa: E501
# @param wildcard wildcard/"*"/String/InputPort/readwrite//False/False/Port receiving the search pattern
# @param filepath filepath/"."/String/InputPort/readwrite//False/False/Port receiving the path to search
# @param file_list file_list//Object.array/OutputPort/readwrite//False/False/Port carrying the resulting file_list
# @par EAGLE_END


class FileGlob(BarrierAppDROP):
    """
    Simple app collecting file names in a directory
    based on a wild-card pattern
    """

    def initialize(self, **kwargs):
        self.wildcard = self._getArg(kwargs, "wildcard", "*")
        self.filepath = self._getArg(kwargs, "filepath", ".")
        BarrierAppDROP.initialize(self, **kwargs)

    def writeData(self):
        """
        Prepare the data and write to all outputs
        """
        for output in self.outputs:
            d = pickle.dumps(self.value)
            output.len = len(d)
            output.write(d)

    def run(self):
        filetmpl = f"{self.filepath}/{self.wildcard}"
        filetmpl = os.path.expandvars(filetmpl)
        logger.debug(f"Looking for files matching {filetmpl}")
        files = glob(filetmpl)
        self.value = [f for f in files if os.path.isfile(f)]
        if len(self.value) == 0:
            logger.warning(f"No matching files found for {filetmpl}")
        else:
            logger.info(
                f"Number of files found matching {filetmpl}: {len(self.value)}"
            )
        self.writeData()


##
# @brief PickOne
# @details App that picks the first element of an input list, passes that
# to all outputs, except the first one. The first output is used to pass
# the remaining array on. This app is useful for a loop.
#
# @par EAGLE_START
# @param category PythonApp
# @param appclass Application Class/dlg_example_cmpts.apps.PickOne/String/ComponentParameter/readonly//False/False/Import path for application class
# @param execution_time Execution Time/5/Float/ComponentParameter/readonly//False/False/Estimated execution time # noqa: E501
# @param num_cpus No. of CPUs/1/Integer/ComponentParameter/readonly//False/False/Number of cores used # noqa: E501
# @param rest_array rest_array//Object.array/InputPort/readwrite//False/FalseList of elements
# @param element element//Object.element/OutputPort/readwrite//False/False/Port carrying the first element of input array
# @param rest_array rest_array//Object.array/OutputPort/readwrite//False/False/Port carrying the rest array
# @par EAGLE_END


class PickOne(BarrierAppDROP):
    """
    Simple app picking one element at a time. Good for Loops.
    """

    def initialize(self, **kwargs):
        BarrierAppDROP.initialize(self, **kwargs)

    def readData(self):
        input = self.inputs[0]
        data = pickle.loads(droputils.allDropContents(input))

        # make sure we always have a ndarray with at least 1dim.
        if type(data) not in (list, tuple) and not isinstance(
            data, (np.ndarray)
        ):
            raise TypeError
        if isinstance(data, np.ndarray) and data.ndim == 0:
            data = np.array([data])
        else:
            data = np.array(data)
        value = data[0] if len(data) else None
        rest = data[1:] if len(data) > 1 else np.array([])
        return value, rest

    def writeData(self, value, rest):
        """
        Prepare the data and write to all outputs
        """
        # write rest to array output
        # and value to every other output
        for output in self.outputs:
            if output.name == "rest_array":
                d = pickle.dumps(rest)
                output.len = len(d)
            else:
                d = pickle.dumps(value)
                output.len = len(d)
            output.write(d)

    def run(self):
        value, rest = self.readData()
        self.writeData(value, rest)


##
# @brief ExtractColumn
# @details App that extracts one column of an table-like ndarray. The array is
# assumed to be row major, i.e. the second index refers to columns.
# The column is # selected by index counting from 0. If the array is 1-D the
# result is a 1 element array. The component is passing whatever type is in
# the selected column.
#
# @par EAGLE_START
# @param category PythonApp
# @param appclass Application Class/dlg_example_cmpts.apps.ExtractColumn/String/ComponentParameter/readonly//False/False/Import path for application class
# @param execution_time Execution Time/5/Float/ComponentParameter/readonly//False/False/Estimated execution time # noqa: E501
# @param num_cpus No. of CPUs/1/Integer/ComponentParameter/readonly//False/False/Number of cores used # noqa: E501
# @param index index/0/Integer/ApplicationArgument/readwrite//False/False/0-base index of column to extract
# @param table_array table_array//array/InputPort/readwrite//False/False/List of elements
# @param column column//Object.1Darray/OutputPort/readwrite//False/False/Port carrying the first element of input array
#               the type is dependent on the list element type.
# @par EAGLE_END
class ExtractColumn(BarrierAppDROP):
    """
    Simple app extracting a column from a 2D ndarray.
    """

    def initialize(self, **kwargs):
        self.index = self._getArg(kwargs, "index", 0)
        BarrierAppDROP.initialize(self, **kwargs)

    def readData(self):
        input = self.inputs[0]
        table_array = pickle.loads(droputils.allDropContents(input))

        # make sure we always have a ndarray with 2dim.
        if not isinstance(table_array, (np.ndarray)) or table_array.ndim != 2:
            raise TypeError
        self.column = table_array[:, self.index]

    def writeData(self):
        """
        Prepare the data and write to all outputs
        """
        # write rest to array output
        # and value to every other output
        for output in self.outputs:
            if output.name == "column":
                d = pickle.dumps(self.column)
                output.len = len(d)
            output.write(d)

    def run(self):
        self.readData()
        self.writeData()


##
# @brief AdvUrlRetrieve
# @details An APP that retrieves the content of a URL and writes
# it to all outputs. The URL can be specified either completely or
# partially through the inputs. In that case the urlTempl parameter can
# use placeholders to construct the final URL.
# @par EAGLE_START
# @param category PythonApp
# @param appclass Application Class/dlg_example_cmpts.apps.AdvUrlRetrieve/String/ComponentParameter/readonly//False/False/Application class
# @param execution_time Execution Time/5/Float/ComponentParameter/readonly//False/False/Estimated execution time # noqa: E501
# @param num_cpus No. of CPUs/1/Integer/ComponentParameter/readonly//False/False/Number of cores used # noqa: E501
# @param urlTempl URL Template/"https://eagle.icrar.org"/String/ApplicationArgument/readwrite//False/False/The URL to retrieve
# @param urlPart URL Part//String/InputPort/readwrite//False/False/The port carrying the content read from the URL.
# @param content Content//String/OutputPort/readwrite//False/False/The port carrying the content read from the URL.
# @par EAGLE_END
class AdvUrlRetrieve(BarrierAppDROP):
    """
    An App that retrieves the content of a URL and allows to construct the URL
    through input placeholders.

    Keywords:
    URL:   string, URL to retrieve.
    """

    def initialize(self, **kwargs):
        self.urlTempl = self._getArg(kwargs, "urlTempl", "")
        BarrierAppDROP.initialize(self, **kwargs)

    def constructUrl(self):
        url = self.urlTempl
        # this will ignore inputs not referenced in template
        for x, v in enumerate(self.urlParts):
            pathRef = "%%i%d" % (x,)
            if pathRef in url:
                url = url.replace(pathRef, v)
        logger.info(f"Constructed URL: {url}")
        return url

    def readData(self):
        # for this app we are expecting URL fractions on the inputs
        urlParts = []
        for input in self.inputs:
            part = pickle.loads(droputils.allDropContents(input))
            # make sure the placeholders are strings
            logger.info(f"URL part: {part}")
            if not isinstance(part, str):
                raise TypeError
            urlParts.append(part)
        self.urlParts = urlParts
        self.url = self.constructUrl()
        try:
            u = urllib.request.urlopen(self.url)
        except Exception as e:
            raise e
        # finally read the content of the URL
        self.content = u.read()

    def writeData(self):
        """
        Prepare the data and write to all outputs
        """
        outs = self.outputs
        written = False
        if len(outs) < 1:
            raise Exception("At least one output required for %r" % self)
        for output in outs:
            if output.name.lower() == "content":
                # we are not pickling here, but just pass on the data.
                output.len = len(self.content)
                output.write(self.content)
                written = True
            else:
                logger.warning(f"Output with name {output.name} ignored!")

        if not written:
            raise TypeError(
                "No matching output drop found." + "Nothing written"
            )

    def run(self):
        self.readData()
        self.writeData()


##
# @brief String2JSON
# @details App that reads a string from a single input and tries
# to convert that to JSON. If port is converted into an argument the
# JSON value is taken from there.
# @par EAGLE_START
# @param category PythonApp
# @param appclass Application Class/dlg_example_cmpts.apps.String2JSON/String/ComponentParameter/readonly//False/False/Application class
# @param execution_time Execution Time/5/Float/ComponentParameter/readonly//False/False/Estimated execution time # noqa: E501
# @param num_cpus No. of CPUs/1/Integer/ComponentParameter/readonly//False/False/Number of cores used # noqa: E501
# @param string string//String/InputPort/readwrite//False/False/String to be converted
# @param element element//Object.json/OutputPort/readwrite//False/False/Port carrying the JSON structure
# @par EAGLE_END
class String2JSON(BarrierAppDROP):

    arg: str = dlg_string_param("string", None)

    def initialize(self, **kwargs):
        BarrierAppDROP.initialize(self, **kwargs)
        if not self.arg:
            self.arg = self._getArg(kwargs, "string", None)

    def readData(self):
        input = self.inputs[0]  # ignore all but the first
        try:
            data = json.loads(droputils.allDropContents(input))
        except json.decoder.JSONDecodeError:
            raise TypeError
        return data

    def writeData(self, data):
        """
        Prepare the data and write to all outputs
        """
        # write rest to array output
        # and value to every other output
        for output in self.outputs:
            d = pickle.dumps(data)
            output.len = len(d)
            output.write(d)

    def run(self):
        if (
            len(self.inputs) == 0
        ):  # if there is no input use the argument value
            try:
                logger.debug("Input found: %s", self.arg)
                data = json.loads(self.arg)
            except TypeError:
                raise
        else:
            data = self.readData()
        self.writeData(data)
        del data  # make sure this is cleaned up


##
# @brief GenericGather
# @details App that reads all its inputs and simply writes them in
# concatenated to all its outputs. This can be used stand-alone or
# as part of a Gather. It does not do anything to the data, just
# passing it on.
#
# @par EAGLE_START
# @param category PythonApp
# @param appclass Application Class/dlg_example_cmpts.apps.GenericGather/String/ComponentParameter/readonly//False/False/Import path for application class
# @param execution_time Execution Time/5/Float/ComponentParameter/readonly//False/False/Estimated execution time # noqa: E501
# @param num_cpus No. of CPUs/1/Integer/ComponentParameter/readonly//False/False/Number of cores used # noqa: E501
# @param input input//Object/InputPort/readwrite//False/False/0-base placeholder port for inputs
# @param output output//Object/OutputPort/readwrite//False/False/Placeholder port for outputs
# @par EAGLE_END
class GenericGather(BarrierAppDROP):
    def initialize(self, **kwargs):
        BarrierAppDROP.initialize(self, **kwargs)

    def readWriteData(self):

        inputs = self.inputs
        outputs = self.outputs
        total_len = 0
        for input in inputs:
            total_len += input.len
        for output in outputs:
            output.len = total_len
            for input in inputs:
                d = droputils.allDropContents(input)
                output.write(d)

    def run(self):
        self.readWriteData()



__package__ = "dlg_example_cmpts"
# The following imports are the binding to the DALiuGE system
# extend the following as required
from .apps import (
    AdvUrlRetrieve,
    ExtractColumn,
    FileGlob,
    GenericGather,
    LengthCheck,
    MyBranch,
    PickOne,
    String2JSON,
)
from .data import MyDataDROP

__all__ = [
    "MyBranch",
    "LengthCheck",
    "MyDataDROP",
    "FileGlob",
    "PickOne",
    "String2JSON",
    "ExtractColumn",
    "AdvUrlRetrieve",
    "GenericGather",
]
